"""
.. module:: get_association_indices
   :synopsis: Determines which indices in the array of STIS associations to
       plot in the preview plots.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
from __future__ import division
from builtins import range
#--------------------
# Package Imports
#--------------------
from spec_plots import __version__

#--------------------
def get_association_indices(associations):
    """
    Given a list of associations, determines the indices that will be plotted.
    If n_associations > 3, this is the first, middle, and last associations,
    otherwise it's all of them.

    :param associations: All of the associations for this STIS spectrum.

    :type associations: list

    :returns: list -- The indices of the list that will be plotted.
    """

    n_associations = len(associations)
    if n_associations <= 3:
        subplot_indices = list(range(n_associations))
    else:
        midindex = int(round(n_associations/2.))
        subplot_indices = [0, midindex, n_associations-1]

    return subplot_indices
#--------------------



"""
.. module:: set_plot_yrange
   :synopsis: Determines the optimal y-axis plot range.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
from builtins import range
#--------------------
# External Imports
#--------------------
import numpy
#--------------------
# Package Imports
#--------------------
from spec_plots import __version__

#--------------------

def set_plot_yrange(wavelengths, fluxes, avoid_regions=None, wl_range=None):
    """
    Given an array of wavelengths, fluxes, and avoid regions, returns a list of
    [ymin,ymax] to define an optimal y-axis plot range.

    :param wavelengths: The wavelengths to be plotted.

    :type wavelengths: numpy.ndarray

    :param fluxes: The fluxes to be plotted.

    :type fluxes: numpy.ndarray

    :param avoid_regions: A list of wavelength ranges to avoid when calculating
        optimal y-axis plot range.

    :type avoid_regions: list of STISAvoidRegion objects.

    :param wl_range: The min. and max. wavelength that defines the x-axis plot
        range.  The default is None, in which case the min. and max. if the
        input wavelength array will be used.

    :type wl_range: list

    :returns: list -- Two-element list containing the optimal [ymin,ymax] values
        to define the y-axis plot range.

    .. note::

         This function makes use of an internal look-up table of wavelength
         regions where known contaminating emission lines or other strong UV
         artifacts can affect the zoom level of the plot.
    """

    # Default the wavelength range to be the entire spectrum if not specified.
    if wl_range is None:
        wl_range = [numpy.nanmin(wavelengths), numpy.nanmax(wavelengths)]

    # This list will keep track of which fluxes to retain when defining the
    # y-axis plot range.  Setting the value to 1 means keep this flux for
    # consideration.
    keep_indices = numpy.asarray([1] * len(wavelengths))

    # Set any fluxes to 0 (don't keep) if they fall within an Avoid Region.
    if avoid_regions is not None:
        for i, region in enumerate(avoid_regions):
            if i == 0:
                # If this is the first Avoid Region, then we need to check if
                # the wavelengths are within the specified bounds supplied
                # through the `wl_range` parameter, in addition to checking if
                # they are within the Avoid Region itself.
                reject_indices = [i for i in range(len(wavelengths)) if (
                    wavelengths[i] >= region.minwl and
                    wavelengths[i] <= region.maxwl or
                    wavelengths[i] < wl_range[0] or
                    wavelengths[i] > wl_range[1])]
            else:
                # Don't need to worry about checking wavelengths within bounds
                # after the first Avoid Region is examined.
                reject_indices = [i for i in range(len(wavelengths)) if (
                    wavelengths[i] >= region.minwl and
                    wavelengths[i] <= region.maxwl)]
            # Set indices that we don't want to keep to 0.  Note that if
            # reject_indices is an empty list then nothing will change.
            keep_indices[reject_indices] = 0

    # After all indices have been set to keep or reject, pull out just the
    # fluxes that should be kept.
    keep_fluxes = numpy.asarray([f for ii, f in enumerate(fluxes) if (
        keep_indices[ii] == 1 and numpy.isfinite(fluxes[ii]))])

    # Don't just take the pure min and max, since large outliers can affect
    # the calculation.  Instead, take the 1th and 99th percentile fluxes within
    # the region to calculate the `min` and `max` fluxes.
    min_flux = numpy.percentile(keep_fluxes, 1.)
    max_flux = numpy.percentile(keep_fluxes, 99.)

    # Determine a y-buffer based on the difference between the max. and min.
    # fluxes.
    ybuffer = 0.3 * (max_flux-min_flux)

    # Make sure the min. and max. fluxes aren't identical (both 0., or both
    # the same exact value.  If so, just return the min. and max. value nudged
    # by 1.0.
    if min_flux != max_flux:
        if min_flux - ybuffer > 0.:
            return [min_flux-ybuffer, max_flux+ybuffer]
        else:
            # We don't want the y-axis range to be TOO far negative, so limit
            # it to be close to the lowest data point. """
            return [1.1*min_flux, max_flux+ybuffer]
    else:
        return [min_flux-1., max_flux+1.]
#--------------------



"""
.. module:: edge_trim
   :synopsis: Returns the start and end indices of a spectrum that represent the
       `best` wavelengths to use when defining the optimal plot axis ranges.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
from __future__ import division
from builtins import zip
#--------------------
# External Imports
#--------------------
import numpy
#--------------------
# Package Imports
#--------------------
from spec_plots.utils.specutils.is_bad_dq import is_bad_dq
from spec_plots import __version__

#--------------------

def _set_plot_xrange_test(instrument, flux_values, flux_err_values, median_flux,
                          flux_scale_factor, median_fluxerr,
                          fluxerr_scale_factor, fluxerr_95th, dqs,
                          check_fluxes=False, check_flux_ratios=False,
                          check_flux_err_ratios=False,
                          check_flux_err_percentile=False, check_dqs=False):
    """
    Defines the test for an invalid part of the spectrum when trimming from the
    edges along the wavelength (x) axis.

    :param instrument: The instrument that is being tested.

    :type instrument: str

    :param flux_values: An array of fluxes to test.

    :type flux_values: numpy.ndarray

    :param flux_err_values: An array of associated flux uncertainties.

    :type flux_err_values: numpy.ndarray

    :param median_flux: A median flux value, used in the test.

    :type median_flux: float

    :param flux_scale_factor: Max. allowed ratio between the flux and a median
        flux value, used in edge trimming.

    :type flux_scale_factor: float

    :param median_fluxerr: A median flux uncertainty value, used in the test.

    :type median_fluxerr: float

    :param fluxerr_scale_factor: Max. allowed ratio between the flux uncertainty
        and a median flux uncertainty value, used in edge trimming.

    :type fluxerr_scale_factor: float

    :param fluxerr_95th: The flux uncertainty corresponding to the 95th
        percentile.

    :type fluxerr_95th: float

    :param dqs: The array of DQ flags to use in the test.  For COS, these are
        the DQ_WGT bits from the header.

    :type dqs: numpy.ndarray

    :param check_fluxes: Should the value of the fluxes be used to test edge
        trimming?

    :type check_fluxes: bool

    :param check_flux_ratios: Should the ratio of the fluxes to the median flux
        value be used to test edge trimming?

    :type check_flux_ratios: bool

    :param check_flux_err_ratios: Should the ratio of the flux uncertainties to
        the median flux uncertainty be used to test edge trimming?

    :type check_flux_err_ratios: bool

    :param check_flux_err_percentile: Should the highest percentile flux
        uncertainties be used to test edge trimming?

    :type check_flux_err_percentile: bool

    :param check_dqs: Should the highest percentile flux uncertainties be used
        to test edge trimming?

    :type check_dqs: bool

    :returns: list -- A list of True/False values depening on whether the input
    flux values pass the test.  Note that if a return value is True, then the
    flux value is considered PART OF THE SPECTRUM TO KEEP.  If median_flux is
    input as NaN, then this function returns False for all flux_values (i.e.,
    skip all of them since median_flux is not defined).

    :raises: ValueError
    """

    # Make sure input parameters are all numpy arrays.
    if (not isinstance(flux_values, numpy.ndarray) or
            not isinstance(flux_err_values, numpy.ndarray) or
            not isinstance(dqs, numpy.ndarray)):
        raise ValueError("The flux, flux uncertainty, and DQ values must be"
                         " passed as a numpy.ndarray object.")

    # Return array of boolean values for the edge_trim test.
    if numpy.isfinite(median_flux):
        bool_results = [((instrument == "cos" and x != 0. and check_fluxes) or
                         (instrument == "stis" and x != 0. and check_fluxes) or
                         (instrument == "miri" and x != 0. and check_fluxes) or
                         (not check_fluxes))
                        and ((abs(x/median_flux) < flux_scale_factor and
                              check_flux_ratios) or (not check_flux_ratios))
                        and ((y/median_fluxerr < fluxerr_scale_factor and
                              check_flux_err_ratios) or
                             (not check_flux_err_ratios))
                        and ((y <= fluxerr_95th and check_flux_err_percentile)
                             or (not check_flux_err_percentile))
                        and ((instrument == "stis" and not
                              is_bad_dq(instrument, z) and check_dqs) or
                             (instrument == "cos" and not
                              is_bad_dq(instrument, z) and check_dqs) or
                             (instrument == "miri" and not
                              is_bad_dq(instrument, z) and check_dqs) or
                             (not check_dqs))
                        for x, y, z in zip(flux_values, flux_err_values, dqs)]
    else:
        bool_results = [False] * len(flux_values)

    return bool_results
#--------------------

#--------------------
def edge_trim(instrument, fluxes, fluxerrs, dqs, n_consecutive, median_flux,
              flux_scale_factor, median_fluxerr, fluxerr_scale_factor,
              fluxerr_95th):
    """
    Returns start and end indexes (end indexes are negatively indexed) of the
    best part of the spectrum to use when defining the plot's wavelength
    ranges.  Returns two sets of start and end indexes: one set without
    taking into account DQ flags, and one set that does take into account DQ
    flags (since some spectra have all DQ flags set > 0).

    :param instrument: The instrument that is being tested.

    :type instrument: str

    :param fluxes: The fluxes to be plotted.

    :type fluxes: numpy.ndarray

    :param fluxerrs: The uncertainties of the fluxes to be plotted.

    :type fluxerrs: numpy.ndarray

    :param dqs: The DQ flags of the spectrum.  For COS, these are the DQ_WGT
        bits from the header.

    :type dqs: numpy.ndarray

    :param n_consecutive: The consecutive number of fluxes that must belong to
        the "best" part of the spectrum for the start/end index to count.

    :type n_consecutive: int

    :param median_flux: The median flux, used in determining where the best part
        of the spectrum is.

    :type median_flux: float

    :param flux_scale_factor: Max. allowed ratio between the flux and a median
        flux value, used in edge trimming.

    :type flux_scale_factor: float

    :param median_fluxerr: The median flux uncertainty, used in determining
        where the best part of the spectrum is.

    :type median_fluxerr: float

    :param fluxerr_scale_factor: Max. allowed ratio between the flux uncertainty
        and a median flux uncertainty value, used in edge trimming.

    :type fluxerr_scale_factor: float

    :param fluxerr_95th: The flux uncertainty corresponding to the 95th
        percentile.

    :type fluxerr_95th: float

    :returns: int tuple -- Indexes that define the best part of the spectrum,
        in the order of (start_index_nodq, end_index_nodq, start_index_withdq,
        end_index_withdq).
    """

    # How many fluxes are there in total?
    n_fluxes = len(fluxes)

    # This defines the size of the spectrum at the edges to examine as a
    # "cheat", without having to look at the entire spectrum to find the good
    # boundaries.  It will look at the whole spectrum as a fallback option.
    edge_size = 100

    # SECTION FOR INDICES IGNORING DQ VALUES

    # First run the xrange test on the first and last 100 elements (for
    # speed).  If no good indices are found within those ranges, then and only
    # then will we run it on the entire array. """
    where_good_fluxes_nodq = numpy.where(numpy.asarray(_set_plot_xrange_test(
        instrument, numpy.concatenate((fluxes[0:edge_size],
                                       fluxes[-1*edge_size:])),
        numpy.concatenate((fluxerrs[0:edge_size],
                           fluxerrs[-1*edge_size:])), median_flux,
        flux_scale_factor, median_fluxerr, fluxerr_scale_factor,
        fluxerr_95th, numpy.concatenate((dqs[0:edge_size], dqs[-1*edge_size:])),
        check_fluxes=True, check_flux_ratios=False,
        check_flux_err_ratios=True, check_flux_err_percentile=False,
        check_dqs=False)))[0]
    if len(where_good_fluxes_nodq) > 0:
        where_good_fluxes_nodq[where_good_fluxes_nodq > edge_size-1] += (
            len(fluxes) - 2*edge_size)
        # Get the start and end indices, ignoring DQ values.
        start_index_nodq, end_index_nodq = find_good_indices(
            where_good_fluxes_nodq, n_consecutive, n_fluxes, first_pass=True)
    else:
        # If there were no results returned from the where call above, make
        # sure the start and end indices will fail the test.
        start_index_nodq = edge_size
        end_index_nodq = -2*edge_size-1

    # If the start and stop indices are not located within the range at the
    # end, run the where on the entire array.
    if (start_index_nodq > edge_size-1 or end_index_nodq < -2*edge_size+
            edge_size+1):
        where_good_fluxes_nodq = numpy.where(numpy.asarray(
            _set_plot_xrange_test(instrument, fluxes, fluxerrs, median_flux,
                                  flux_scale_factor, median_fluxerr,
                                  fluxerr_scale_factor, fluxerr_95th, dqs,
                                  check_fluxes=True,
                                  check_flux_ratios=False,
                                  check_flux_err_ratios=True,
                                  check_flux_err_percentile=False,
                                  check_dqs=False)))[0]
        start_index_nodq, end_index_nodq = find_good_indices(
            where_good_fluxes_nodq, n_consecutive, n_fluxes)

    # SECTION FOR INDICES INCLUDING DQ VALUES

    # First run the xrange test on the first and last 100 elements (for
    # speed).  If no good indices are found within those ranges, then and only
    # then will we run it on the entire array.
    where_good_fluxes_withdq = numpy.where(numpy.asarray(_set_plot_xrange_test(
        instrument, numpy.concatenate((fluxes[0:edge_size],
                                       fluxes[-1*edge_size:])),
        numpy.concatenate((fluxerrs[0:edge_size],
                           fluxerrs[-1*edge_size:])), median_flux,
        flux_scale_factor, median_fluxerr, fluxerr_scale_factor,
        fluxerr_95th, numpy.concatenate((dqs[0:edge_size],
                                         dqs[-1*edge_size:])),
        check_fluxes=True, check_flux_ratios=False,
        check_flux_err_ratios=True, check_flux_err_percentile=False,
        check_dqs=True)))[0]
    if len(where_good_fluxes_withdq) > 0:
        where_good_fluxes_withdq[where_good_fluxes_withdq > edge_size-1] += (
            len(fluxes) - 2*edge_size)
        # Get the start and end indices, including DQ values.
        start_index_withdq, end_index_withdq = find_good_indices(
            where_good_fluxes_withdq, n_consecutive, n_fluxes, first_pass=True)
    else:
        # If there were no results returned from the where call above, make
        # sure the start and end indices will fail the test.
        start_index_withdq = edge_size
        end_index_withdq = -2*edge_size-1

    # If the start and stop indices are not located within the range at the
    # end, run the where on the entire array.
    if (start_index_withdq > edge_size-1 or end_index_withdq < -2*edge_size+
            edge_size+1):
        where_good_fluxes_withdq = numpy.where(numpy.asarray(
            _set_plot_xrange_test(instrument, fluxes, fluxerrs, median_flux,
                                  flux_scale_factor, median_fluxerr,
                                  fluxerr_scale_factor, fluxerr_95th, dqs,
                                  check_fluxes=True,
                                  check_flux_ratios=False,
                                  check_flux_err_ratios=True,
                                  check_flux_err_percentile=False,
                                  check_dqs=True)))[0]
        start_index_withdq, end_index_withdq = find_good_indices(
            where_good_fluxes_withdq, n_consecutive, n_fluxes)

    return (start_index_nodq, end_index_nodq, start_index_withdq,
            end_index_withdq)
#--------------------

#--------------------
def find_good_indices(indices_arr, n_consecutive, n_fluxes, first_pass=False):
    """
    Given an array of indices, determines the start and stop indices of the
    `best` part of the spectrum based on the required number of consecutive
    `good` fluxes.

    :param indices_arr: Something

    :type indices_arr: list

    :param n_consecutive: The consecutive number of fluxes that must belong to
        the "best" part of the spectrum for the start/end index to count.

    :type n_consecutive: int

    :param n_fluxes: The total number of data points in the spectrum.

    :type n_fluxes: int

    :param first_pass: Used to determine if this is the first attempt to find
        good start and stop indices or not.  The first pass normally uses only a
        part of the spectrum (pieces at the beginning and end), while the next
        pass uses the entire spectrum.  If it's not the first pass and no good
        start and stop indices are found, then it must return the last and first
        index.

    :type first_pass: bool

    :returns: int tuple -- Indices that define the best part of the spectrum
        (start and stop).
    """

    # How many total good indices are there to work with?
    n_good_fluxes = len(indices_arr)

    # Handle the trivial case where the number of valid points is less than
    # the `n_consecutive` desired.  In this case, the start and end indices are
    # set to the last and first indices, simulating what would happen if the
    # entire array was traversed.
    if n_good_fluxes < n_consecutive:
        if not first_pass:
            start_index = 0
            end_index = -1
        else:
            start_index = n_fluxes
            end_index = -1 * n_fluxes
    else:
        # The start index is then the min. index for which the
        # `n_consecutive`'th index is exactly equal to `n_consecutive`-1.
        start_indexes_with_good_diffs = numpy.asarray(
            [x for i, x in enumerate(indices_arr[0:n_good_fluxes-n_consecutive+
                                                 1]) if
             indices_arr[i+n_consecutive-1]-indices_arr[i] == n_consecutive-1])

        if len(start_indexes_with_good_diffs) > 0:
            start_index = numpy.min(start_indexes_with_good_diffs)
            end_index = (numpy.max(start_indexes_with_good_diffs) - n_fluxes +
                         n_consecutive - 1)
        else:
            if not first_pass:
                start_index = 0
                end_index = -1
            else:
                start_index = n_fluxes
                end_index = -1 * n_fluxes

    return start_index, end_index
#--------------------



"""
.. module:: debug_oplot
   :synopsis: Plots spectral data points color-coded based on a suite of
       rejection criteria, primarily for debugging purposes.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
from __future__ import division
#--------------------
# External Imports
#--------------------
import numpy
#--------------------
# Package Imports
#--------------------
from spec_plots.utils.specutils.is_bad_dq import is_bad_dq
from spec_plots import __version__

#--------------------

def debug_oplot(this_plotarea, instrument, all_wls, all_fls, all_flerrs,
                all_dqs, median_flux, median_fluxerr, flux_scale_factor,
                fluxerr_scale_factor, fluxerr_95th, oplot_percentiles=False):
    """
    Creates plots of the spectra with color-coding and special annotation to
    identify which points were rejected by which tests.  Useful for
    debugging and understanding why a given plot had its plot axes defined
    the way it did.

    :param this_plotarea: The AxesSubplot object to plot on.

    :type this_plotarea: matplotlib.axes._subplots.AxesSubplot

    :param instrument: The instrument that is being tested.

    :type instrument: str

    :param all_wls: Array of wavelengths.

    :type all_wls: numpy.ndarray

    :param all_fls: Array of fluxes.

    :type all_fls: numpy.ndarray

    :param all_flerrs: Array of flux uncertainties.

    :type all_flerrs: numpy.ndarray

    :param all_dqs: Array of data quality flags.  For COS, these are the DQ_WGT
        bits from the header.

    :type all_dqs: numpy.ndarray

    :param median_flux: The median flux used in determining where the best part
        of the spectrum is.

    :type median_flux: float

    :param median_fluxerr: The median flux uncertainty used in determining where
        the best part of the spectrum is.

    :type median_fluxerr: float

    :param flux_scale_factor: Max. allowed ratio between the flux and a median
        flux value used in edge trimming.

    :type flux_scale_factor: float

    :param fluxerr_scale_factor: Max. allowed ratio between the flux uncertainty
        and a median flux uncertainty value used in edge trimming.

    :type fluxerr_scale_factor: float

    :param fluxerr_95th: The flux uncertainty corresponding to the 95th
        percentile.

    :type fluxerr_95th: float

    :param oplot_percentiles: Set this to True to overplot points where the flux
        uncertainties are greater than the 95th percentile.  Default = False.

    :type oplot_percentiles: bool
    """

    # Plot all of the points in black.
    this_plotarea.errorbar(all_wls, all_fls, yerr=all_flerrs, ecolor='c',
                           color='k', label='Passed')

    # Plot those fluxes that fail because they are much greater than the median
    # in blue.
    if numpy.isfinite(median_flux):
        where_fluxtoolarge = numpy.where(abs(all_fls/median_flux) >
                                         flux_scale_factor)
        if len(where_fluxtoolarge[0]) > 0:
            this_plotarea.plot(all_wls[where_fluxtoolarge],
                               all_fls[where_fluxtoolarge], 'bo',
                               label="Flux>>Median")

    # Plot those fluxes that fail because they are exactly equal to 0 in green.
    where_allzero = numpy.where(all_fls == 0.0)
    if len(where_allzero[0]) > 0:
        this_plotarea.plot(all_wls[where_allzero], all_fls[where_allzero],
                           'go', label="Flux=0")

    # Plot those fluxes that fail because they have a bad DQ value in red.
    where_bad_dq = numpy.where(is_bad_dq(instrument, all_dqs))[0]
    if len(where_bad_dq) > 0:
        this_plotarea.plot(all_wls[where_bad_dq], all_fls[where_bad_dq], 'ro',
                           label="BAD DQ_WGT")

    # Plot those fluxes that fail because their flux uncertainties are much
    # greater than the median flux uncertainty in magenta.
    if numpy.isfinite(median_fluxerr):
        where_bigerr = numpy.where(all_flerrs/median_fluxerr >
                                   fluxerr_scale_factor)
        if len(where_bigerr[0]) > 0:
            this_plotarea.plot(all_wls[where_bigerr], all_fls[where_bigerr],
                               'mo', label="FluxErr>>Median")

    # Plot those fluxes that fail because their flux uncertainties are greater
    # than the 95th percentile in yellow.
    if oplot_percentiles:
        where_bigerrpercentile = numpy.where(all_flerrs > fluxerr_95th)
        if len(where_bigerrpercentile[0]) > 0:
            this_plotarea.plot(all_wls[where_bigerrpercentile],
                               all_fls[where_bigerrpercentile], 'yo',
                               label="FluxErr>95th %")

    this_plotarea.legend(loc="upper center", ncol=4)
#--------------------



"""
.. module:: check_segments
   :synopsis: Checks that the array of segment names in the COS FITS extension
       table are some combination of allowed values.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
import os
import sys
from builtins import str
#--------------------
# External Imports
#--------------------
import numpy
#--------------------
# Package Imports
#--------------------
from spec_plots.utils.specutils.specutilserror import SpecUtilsError
from spec_plots import __version__
#--------------------

#--------------------
def check_segments(segments_list, input_file):
    """
    Checks that the array of "segments" in the COS spectrum header are expected
    values.  It returns a scalar string representing the band (either "FUV"
    or "NUV").  If the array is not what's expected, an Exception is raised.

    :param segments_list: List of segment labels.

    :type segments_list: list

    :param input_file: Name of input FITS file.

    :type input_file: str

    :returns: str -- A string representation of the band, either "FUV" or "NUV".

    :raises: utils.specutils.SpecUtilsError

    .. note::

         This function does not attempt to access the input file, it only
         requires the name of the file for error reporting purposes.
    """

    # Get number of segments.
    n_segments = len(segments_list)

    # Segment array must contain either one, two, or three elements.
    if n_segments == 1:
        # This can happen with FUV data for some reason, so it can be either
        # "FUVA" or "FUVB".
        if (numpy.array_equal(segments_list, ["FUVA"]) or
                numpy.array_equal(segments_list, ["FUVB"])):
            this_band = "FUV"
        else:
            raise SpecUtilsError("The array of SEGMENT strings contains one"
                                 " value, but is not equal to [\"FUVA\"] or"
                                 " [\"FUVB\"] in file " + input_file)

    elif n_segments == 2:
        # Must be ["FUVA", "FUVB"].
        if numpy.array_equal(segments_list, ["FUVA", "FUVB"]):
            this_band = "FUV"
        else:
            raise SpecUtilsError("The array of SEGMENT strings contains two"
                                 " values, but is not equal to [\"FUVA\", "
                                 "\"FUVB\"] in file " + input_file)

    elif n_segments == 3:
        # Must be ["NUVA", "NUVB", "NUVC"].
        if numpy.array_equal(segments_list, ["NUVA", "NUVB", "NUVC"]):
            this_band = "NUV"
        else:
            raise SpecUtilsError("The array of SEGMENT strings contains three"
                                 " values, but is not equal to [\"NUVA\","
                                 " \"NUVB\", \"NUVC\"] in file " + input_file)

    else:
        raise SpecUtilsError("The array of SEGMENT strings should contain 1,"
                             " 2, or 3 values, found " + str(n_segments) +
                             " in file " + input_file)

    return this_band
#--------------------



"""
.. module:: __init__

   :synopsis: Used to treat "specutils_cos" as a directory and package.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
#--------------------
# Package Imports
#--------------------
from spec_plots import __version__
from spec_plots.utils.specutils_cos.cosspectrum import COSSpectrum
from spec_plots.utils.specutils_cos.check_segments import check_segments
from spec_plots.utils.specutils_cos.extract_subspec import extract_subspec
from spec_plots.utils.specutils_cos.plotspec import plotspec
from spec_plots.utils.specutils_cos.readspec import readspec
from spec_plots.utils.specutils_cos.get_segment_names import get_segment_names



"""
.. module:: readspec
   :synopsis: Reads in a STIS spectrum from a FITS file.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
from builtins import range
#--------------------
# External Imports
#--------------------
from astropy.io import fits
#--------------------
# Package Imports
#--------------------
from spec_plots.utils.specutils_stis.stis1dspectrum import (
    STIS1DSpectrum, STISExposureSpectrum, STISOrderSpectrum)
from spec_plots import __version__

#--------------------

#--------------------
def readspec(input_file):
    """
    Reads in a STIS spectrum FITS file (x1d, sx1) and returns the wavelengths,
    fluxes, and flux uncertainties for the two (FUV segments) or three (NUV
    stripes).

    :param input_file: Name of input FITS file.

    :type input_file: str

    :returns: STIS1DSpectrum -- The spectroscopic data (wavelength, flux, flux
        error, etc):
    """

    with fits.open(input_file) as hdulist:
        # Create an initially empty list that will contain each extension's
        # (association's) spectrum object.
        all_association_spectra = []

        # Loop over each association and create the COS spectrum objects.
        for exten in hdulist[1:]:
            exten_data_table = exten.data

            # How many orders (table rows) in this extension?
            n_orders = len(exten_data_table["sporder"])

            # Create a list of STISOrderSpectra for this extension.
            all_order_spectra = [STISOrderSpectrum(
                nelem=exten_data_table["nelem"][order],
                wavelengths=exten_data_table["WAVELENGTH"][order],
                fluxes=exten_data_table["FLUX"][order],
                fluxerrs=exten_data_table["ERROR"][order],
                dqs=exten_data_table["DQ"][order]) for order in range(n_orders)]

            # Create a STISExposureSpectrum from the STISOrderSpectrum
            # objects.  Append to the running list of them.
            this_exposure_spectrum = STISExposureSpectrum(all_order_spectra)
            all_association_spectra.append(this_exposure_spectrum)

        return STIS1DSpectrum(all_association_spectra, orig_file=input_file)
#--------------------



"""
.. module:: specutilserror
   :synopsis: Defines a custom Error class for use with spec_plots utility
       functions.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
#--------------------
# Package Imports
#--------------------
from spec_plots import __version__

#--------------------

class SpecUtilsError(Exception, object):
    """
    This class defines a generic Exception to use for errors raised in the
    SPECUTILS modules (specutils, specutils_cos, specutils_stis, etc.).  It
    simply prints the given value when raising the exception, e.g.,

    .. code-block:: python

         raise SpecUtilsError("Print this string")
         SpecUtilsError: *** SPECUTILS ERROR: 'Print this string'
    """

    def __init__(self, value):
        """
        Initiate the Exception.

        :param value: The string to print on error.

        :type value: str
        """
        super(SpecUtilsError, self).__init__(value)
        self.value = value

    def __str__(self):
        """
        Overrides the str function for this class.
        """
        return "*** SPECUTILS ERROR: "+repr(self.value)
#--------------------



"""
.. module:: plotspec
   :synopsis: Creates preview plots for the provided STIS spectrum.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
from __future__ import division
import copy
import os
import sys
from builtins import str
from builtins import range
#--------------------
# External Imports
#--------------------
import matplotlib
from matplotlib.ticker import FormatStrFormatter
import matplotlib.pyplot as pyplot
from matplotlib import rc
import numpy
#--------------------
# Package Imports
#--------------------
from spec_plots.utils.specutils.specutilserror import SpecUtilsError
from spec_plots.utils.specutils.debug_oplot import debug_oplot
from spec_plots.utils.specutils.calc_covering_fraction import (
    calc_covering_fraction)
from spec_plots import __version__

if matplotlib.get_backend().lower() != 'agg':
    pyplot.switch_backend('Agg')
#--------------------

#--------------------
def plotspec(stis_spectrum, association_indices, stitched_spectra, output_type,
             output_file, flux_scale_factor,
             fluxerr_scale_factor, plot_metrics, dpi_val=96., output_size=1024,
             debug=False, full_ylabels=False, optimize=True):
    """
    Accepts a STIS1DSpectrum object from the READSPEC function and produces
    preview plots.

    :param stis_spectrum: STIS spectrum as returned by READSPEC.

    :type stis_spectrum: STIS1DSpectrum

    :param association_indices: The indices of the original list of associations
        for this spectrum that will be plotted.  Used in the informational plot
        title, e.g., `Association <x> of <y>.`, where <x> is the association
        index and <y> is the total number of associations in this spectrum.

    :type association_indices: list

    :param stitched_spectra: List of order-stitched spectra for each
        association, as created in `utils.specutils.stitch_components()`.

    :type stitched_spectra: list

    :param output_type: What kind of output to make?

    :type output_type: str

    :param output_file: Name of output file (including full path).

    :type output_file: str

    :param flux_scale_factor: Max. allowed ratio between the flux and a median
        flux value, used in edge trimming.  Default = 10.

    :type flux_scale_factor: float

    :param fluxerr_scale_factor: Max. allowed ratio between the flux uncertainty
        and a median flux uncertainty value, used in edge trimming.
        Default = 5.

    :type fluxerr_scale_factor: float

    :param plot_metrics: Collection of plot metrics (flux statistics, axis
        ranges, etc.) to use when making the plots.  These are computed using
        `utils.specutils.calc_plot_metrics()`.

    :type plot_metrics: list

    :param dpi_val: The DPI value of your device's monitor.  Affects the size of
        the output plots.  Default = 96. (applicable to most modern monitors).

    :type dpi_val: float

    :param output_size: Size of plot in pixels (plots are square in dimensions).
        Defaults to 1024.

    :param output_size: int

    :param debug: Should the output plots include debugging information
        (color-coded data points based on rejection criteria, shaded exclude
        regions)?  Default = False.

    :type debug: bool

    :param full_ylabels: Should the y-labels contain the full values (including
        the power of 10 in scientific notation)?  Default = False.

    :type full_ylabels: bool

    :param optimize: If set to True, will use a slightly optimized version of
        determining the plot covering fraction.

    :type optimize: bool

    :raises: OSError, utils.specutils.SpecUtilsError

    .. note::

         This function assumes a screen resolution of 96 DPI in order to
         generate plots of the desired sizes.  This is because matplotlib works
         in units of inches and DPI rather than pixels.
    """

    # Make sure the plot size is set to an integer value.
    if output_size is not None:
        if not isinstance(output_size, int):
            output_size = int(round(output_size))

    # Make sure the output path exists, if not, create it.
    if output_type != 'screen':
        if (os.path.dirname(output_file) != "" and
                not os.path.isdir(os.path.dirname(output_file))):
            try:
                os.mkdir(os.path.dirname(output_file))
            except OSError as this_error:
                if this_error.errno == 13:
                    sys.stderr.write("*** MAKE_HST_SPEC_PREVIEWS ERROR:"
                                     " Output directory could not be created,"
                                     " "+repr(this_error.strerror)+"\n")
                    exit(1)
                else:
                    raise

    # If the figure size is large, then plot up to three associations,
    # otherwise force only one association on the plot.
    n_associations = len(stis_spectrum.associations)
    n_subplots = len(stitched_spectra)
    is_bigplot = output_size > 128

    # Start plot figure.
    this_figure, these_plotareas = pyplot.subplots(nrows=n_subplots, ncols=1,
                                                   figsize=(
                                                       output_size/dpi_val,
                                                       output_size/dpi_val),
                                                   dpi=dpi_val)

    # Make sure the subplots are in a numpy array (I think by default it is
    # not if there is only one).
    if not isinstance(these_plotareas, numpy.ndarray):
        these_plotareas = numpy.asarray([these_plotareas])

    # Adjust the plot geometry (margins, etc.) based on plot size.
    if is_bigplot:
        this_figure.subplots_adjust(hspace=0.3, top=0.915)
    else:
        this_figure.subplots_adjust(top=0.85, bottom=0.3, left=0.25, right=0.8)

    # Loop over each association.
    covering_fractions = [0.] * len(stitched_spectra)
    for i in range(len(stitched_spectra)):
        this_plotarea = these_plotareas[i]

        # Get the wavelengths, fluxes, flux uncertainties, and data quality
        # flags out of the stitched spectrum for this association.
        try:
            all_wls = stitched_spectra[i]["wls"]
            all_fls = stitched_spectra[i]["fls"]
            all_flerrs = stitched_spectra[i]["flerrs"]
            all_dqs = stitched_spectra[i]["dqs"]
            title_addendum = stitched_spectra[i]["title"]
        except KeyError as the_error:
            raise SpecUtilsError("The provided stitched spectrum does not have"
                                 " the expected format, missing key " +
                                 str(the_error)+".")

        # Only plot information in the plot title if the plot is large (and
        # therefore sufficient space exists on the plot).
        if is_bigplot:
            this_plotarea.set_title(title_addendum, loc="right", size="small",
                                    color="red")
        if n_associations > 1 and is_bigplot:
            this_plotarea.set_title("Association "+
                                    str(association_indices[i]+1)+"/"+
                                    str(n_associations), loc="center",
                                    size="small", color="black")

        # Extract the optimal x-axis plot range from the plot_metrics dict,
        # since we use it a lot.
        optimal_xaxis_range = plot_metrics[i]["optimal_xaxis_range"]

        # Plot the spectrum, but only if valid wavelength ranges for x-axis
        # are returned, otherwise plot a special "Fluxes Are All Zero" plot.
        if all(numpy.isfinite(optimal_xaxis_range)):
            # We plot the spectrum as a regular line for use
            # in calc_covering_fraction, it will be removed later.
            this_line = this_plotarea.plot(all_wls, all_fls, 'b')

            # Update y-axis range, but only adjust the ranges if this isn't
            # an all-zero flux case (and not in debug mode, in which case I want
            # to see the entire y-axis range).
            if not debug:
                this_plotarea.set_ylim(plot_metrics[i]["y_axis_range"])

            covering_fractions[i] = calc_covering_fraction(this_figure,
                                                           these_plotareas, i,
                                                           optimize=optimize)
            # Note: here we remove the line we plotted before, it was only
            # so that calc_covering_fraction would have someting to draw on the
            # canvas and thereby determine which pixels were "blue" (i.e., part
            # of the plotted spectrum vs. background).
            this_plotarea.lines.remove(this_line[0])
            # Now we plot the spectrum as a LineCollection so that the
            # transparency will have the desired effect, but, this is not
            # rendered on the canvas inside calc_covering_fraction, hence why we
            # need to plot it both as a regular line first.
            # Note: because we are re-using a LineCollection object in the
            # array of plot_metrics (specifically, when creating the
            # thumbnail-sized plot), we have to use a copy of the LineCollection
            # object, otherwise it will have Axes, Figure, etc. all defined and
            # resetting them to None does not work.  Since this is only an issue
            # with thumbnail-sizes, this is only relevant for the first
            # LineCollection.
            if i == 0:
                this_collection = this_plotarea.add_collection(copy.copy(
                    plot_metrics[i]["line_collection"]))
            else:
                this_collection = this_plotarea.add_collection(
                    plot_metrics[i]["line_collection"])

            if covering_fractions[i] > 30.:
                this_collection.set_alpha(0.1)

            # Turn on plot grid lines.
            this_plotarea.grid(True)

            if is_bigplot:
                if i == len(stitched_spectra)-1:
                    this_figure.suptitle(os.path.basename(
                        stis_spectrum.orig_file), fontsize=18, color='r')
                else:
                    this_figure.suptitle(os.path.basename(
                        stis_spectrum.orig_file), fontsize=18,
                                         color='white')
                # Uncomment the lines below to include the covering fraction
                # as part of the suptitle.
#                if i == len(stitched_spectra)-1:
#                    this_figure.suptitle(os.path.basename(
#                            stis_spectrum.orig_file) + ": " +
#                                         ','.join(['{0:6.2f}'.format(y)
#                                                   for y in covering_fractions]
#                                                  ), fontsize=18, color='k')
#                else:
#                    this_figure.suptitle(os.path.basename(
#                            stis_spectrum.orig_file) + ": " +
#                                         ','.join(['{0:6.2f}'.format(y)
#                                                   for y in covering_fractions]
#                                                  ), fontsize=18,
#                                         color='white')

            if debug:
                # Overplot points color-coded based on rejection criteria.
                debug_oplot(this_plotarea, "stis", all_wls, all_fls,
                            all_flerrs, all_dqs,
                            plot_metrics[i]["median_flux"],
                            plot_metrics[i]["median_fluxerr"],
                            flux_scale_factor, fluxerr_scale_factor,
                            plot_metrics[i]["fluxerr_95th"])

                # Overplot the x-axis edges that are trimmed to define the
                # y-axis plot range as a shaded area.
                this_plotarea.axvspan(numpy.nanmin(all_wls),
                                      optimal_xaxis_range[0],
                                      facecolor="lightgrey", alpha=0.5)
                this_plotarea.axvspan(optimal_xaxis_range[1],
                                      numpy.nanmax(all_wls),
                                      facecolor="lightgrey", alpha=0.5)

                # Overplot the avoid regions in a light color as a shaded area.
                for region in plot_metrics[i]["avoid_regions"]:
                    this_plotarea.axvspan(region.minwl, region.maxwl,
                                          facecolor="lightgrey", alpha=0.5)

            # This is where we ensure the x-axis range is based on the full
            # x-axis range, rather than using the optimum x-axis range.  This is
            # done so that all the plots for a similar instrument setting will
            # have the same starting and ending plot values.
            min_wl = numpy.nanmin(all_wls)
            max_wl = numpy.nanmax(all_wls)
            xplot_buffer = (max_wl - min_wl) * 0.05
            this_plotarea.set_xlim([min_wl-xplot_buffer, max_wl+xplot_buffer])

            # Only use two tick labels (min and max wavelengths) for
            # thumbnails, because there isn't enough space otherwise.
            if not is_bigplot:
                rc('font', size=10)
                minwl = numpy.nanmin(all_wls)
                maxwl = numpy.nanmax(all_wls)
                this_plotarea.set_xticks([minwl, maxwl])
                this_plotarea.set_xticklabels(this_plotarea.get_xticks(),
                                              rotation=25.)
                this_plotarea.xaxis.set_major_formatter(FormatStrFormatter(
                    "%6.1f"))
            else:
                # Make sure the font properties go back to normal.
                pyplot.rcdefaults()
                this_plotarea.tick_params(axis='x', labelsize=14)
                this_plotarea.set_xlabel(r"Wavelength $(\AA)$", fontsize=16,
                                         color='k')
                this_plotarea.set_ylabel(r"Flux $\mathrm{(erg/s/cm^2\!/\AA)}$",
                                         fontsize=16, color='k')

                # If requested, include the powers of 10 part of the y-axis
                # tickmarks.
                if full_ylabels:
                    this_plotarea.yaxis.set_major_formatter(FormatStrFormatter(
                        '%3.2E'))

        else:
            # Otherwise this is a spectrum that has all zero fluxes, or some
            # other problem, and we make a default plot.  Define the optimal
            # x-axis range to span the original spectrum.
            optimal_xaxis_range = [numpy.nanmin(all_wls), numpy.nanmax(all_wls)]
            this_plotarea.set_xlim(optimal_xaxis_range)

            # Make the plot background grey to distinguish that this is a
            # `special` plot.  Turn off y-tick labels.
            this_plotarea.set_axis_bgcolor("lightgrey")
            this_plotarea.set_yticklabels([])

            # Configure the plot units, text size, and other markings based
            # on whether this is a large or thumbnail-sized plot.
            if not is_bigplot:
                rc('font', size=10)
                minwl = numpy.nanmin(all_wls)
                maxwl = numpy.nanmax(all_wls)
                this_plotarea.set_xticks([minwl, maxwl])
                this_plotarea.set_xticklabels(this_plotarea.get_xticks(),
                                              rotation=25.)
                this_plotarea.xaxis.set_major_formatter(FormatStrFormatter(
                    "%6.1f"))
                textsize = "small"
                plottext = "Fluxes are \n all 0."
            else:
                # Make sure the font properties go back to normal.
                pyplot.rcdefaults()
                this_plotarea.tick_params(axis='x', labelsize=14)
                this_plotarea.set_xlabel(r"Wavelength $(\AA)$", fontsize=16,
                                         color='k')
                this_plotarea.set_ylabel(r"Flux $\mathrm{(erg/s/cm^2\!/\AA)}$",
                                         fontsize=16, color='k')

                # If requested, include the powers of 10 part of the y-axis
                # tickmarks.
                if full_ylabels:
                    this_plotarea.yaxis.set_major_formatter(FormatStrFormatter(
                        '%3.2E'))

                textsize = "x-large"
                plottext = "Fluxes are all 0."

            # Place the text with the informational message in the center of
            # the plot.
            this_plotarea.text(0.5, 0.5, plottext, horizontalalignment="center",
                               verticalalignment="center",
                               transform=this_plotarea.transAxes, size=textsize)

    # Display or plot to the desired format.
    if output_type != "screen":
        # Deconstruct output file to include plot size information.
        output_splits = os.path.split(output_file)
        file_splits = os.path.splitext(output_splits[1])
        if output_splits[0] != "":
            revised_output_file = (output_splits[0]+os.path.sep+file_splits[0] +
                                   '_{0:04d}'.format(output_size) +
                                   file_splits[1])
        else:
            revised_output_file = (file_splits[0] +
                                   '_{0:04d}'.format(output_size) +
                                   file_splits[1])

        # Save figure.
        this_figure.savefig(revised_output_file, format=output_type,
                            dpi=dpi_val)

    elif output_type == "screen":
        pyplot.show()
#--------------------



"""
.. module:: make_html
   :synopsis: Creates a webpage of thumbnail and full-size previews for review
       purposes.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
from __future__ import print_function
import argparse
from glob import glob
import os
import sys
from builtins import str
#--------------------
# External Imports
#--------------------
import numpy
#--------------------
# Package Imports
#--------------------
from spec_plots import __version__

#--------------------

def find_orig_preview(ufr, orig_files):
    """
    Given aunique file root (`ufr`, which includes the IPPPSSOOT ID), this
    function matches that to the list of original preview files from CADC to
    determine if an original preview exists for this ufr.

    :param ufr: Unique file root to test (this is equivalent to the IPPPSSOOT ID
        from MAST).

    :type ufr: str

    :param orig_files: Array of original preview file names in which to find a
        match.

    :type orig_files: numpy.ndarray

    :returns: int -- The index in `orig_files` that matches the specified `ufr`,
        if there is no match then returns None.
    """

    # Find all indices where the ufr is a substring of the original CADC
    # preview plot file name.  Note that the ufr is input as <ufr>_<size>, so
    # split on underscore and just look for the first part.
    where_match = [i for i, x in enumerate(orig_files) if
                   ufr.split('_')[0] in x]

    # Return the index if found.
    if len(where_match) == 1:
        return where_match[0]
    elif len(where_match) > 1:
        print("*** WARNING in MAKE_HTML: Found more than one CADC match, using"
              " the first one for file " + ufr)
        return where_match[0]
    else:
        return None

#--------------------

def make_html(idir, odir="html/plot_previews/", ofile="plot_previews",
              orig_dir=None, plot_display_width=512):
    """
    Creates diagnostic HTML page showing the generated preview plots (thumb and
    large) as well as the original MAST version of these plots (the one's from
    CADC, provided they can be located in the expected location).

    :param idir: Input directory containing the set of preview images generated
        by spec_plots.

    :type idir: str

    :param odir: Output path for the HTML files.  Default = html/plot_previews/`

    :type odir: str

    :param ofile: Base file name of the output HTML files.  Default =
        plot_previews`

    :type ofile: str

    :param orig_dir: [Optional] Directory containing the original preview plots
        from CADC.  If provided, these will be included in the HTML pages for
        comparison purposes.

    :type orig_dir: str

    :param plot_display_width: [Optional] The width of the preview plots in the
        HTML tables (specified through the <img> tag), in pixels.  Default =
        512.

    :type plot_display_width: int

    :raises: OSError, IOError, ValueError
    """

    # Ensure output directory is treated as an absolute path, make sure it
    # exists, if not, create the output directory.
    odir = os.path.abspath(odir) + os.path.sep
    if odir != '' and not os.path.isdir(odir):
        try:
            os.mkdir(odir)
        except OSError as this_error:
            if this_error.errno == 13:
                sys.stderr.write("*** MAKE_HTML ERROR: Output directory could"
                                 " not be created, "+repr(this_error.strerror)+
                                 "\n")
                exit(1)
            else:
                raise

    # Get all of the thumb and full-size .png files in this directory.
    if idir is not None:
        if os.path.isdir(idir):
            all_thumb_png_files = numpy.asarray(glob(idir+"/*0128.png"))
            n_thumb_png_files = len(all_thumb_png_files)

            all_large_png_files = numpy.asarray(glob(idir+"/*1024.png"))
            n_large_png_files = len(all_large_png_files)

            # <DEVEL> The code blocks that create the file roots should be
            # generators and not arrays, since the array size might be quite
            # large. </DEVEL>
            # Get the file root of the thumb-size spec_plots preview images,
            # which are the first two parts of the file name separated by
            # underscores.  E.g., la7803fkq_x1d_0128.png -> la7803fkq_x1d.
            if n_thumb_png_files > 0:
                all_thumb_png_files_froots = numpy.asarray(
                    ['_'.join(os.path.basename(x).split('_')[0:2])
                     for x in all_thumb_png_files])
                thumb_sort_indexes = numpy.argsort(all_thumb_png_files_froots)
                all_thumb_png_files_froots = (
                    all_thumb_png_files_froots[thumb_sort_indexes])
                all_thumb_png_files = all_thumb_png_files[thumb_sort_indexes]


            # Get the file root for the large-size spec_plots preview files.
            if n_large_png_files > 0:
                all_large_png_files_froots = numpy.asarray(
                    ['_'.join(os.path.basename(x).split('_')[0:2])
                     for x in all_large_png_files])
                large_sort_indexes = numpy.argsort(all_large_png_files_froots)
                all_large_png_files_froots = (
                    all_large_png_files_froots[large_sort_indexes])
                all_large_png_files = all_large_png_files[large_sort_indexes]

        else:
            raise IOError("Could not find directory " + idir)

        # If requested, get a list of all the original preview plots.
        if orig_dir is not None:
            if os.path.isdir(orig_dir):
                all_orig_files_withpath = numpy.asarray(glob(orig_dir+"/*.jpg"))
                all_orig_files = numpy.asarray(
                    [os.path.basename(x).lower() for x in
                     all_orig_files_withpath])
            else:
                raise IOError("Could not find directory containing original"
                              " preview plots.  Looking for " + orig_dir)

        # Get a list of unique <IPPPSSOOT_filetype> base names from both the
        # all_thumb and all_large arrays of spec_plots file names.  This is
        # because, in principle, there might only be a preview thumbnail or a
        # full-size preview for a given IPPPSSOOT_filetype.  Note that this
        # array should be sorted since the "unique" function returns a sorted
        #  array.
        # <DEVEL> Concatenating the arrays here can be expensive if trying to
        # make webpage for a very large number of preview plots.
        # Consider an alternative way of getting the unique set between the two
        # without creating a new array. </DEVEL>
        if n_thumb_png_files > 0 and n_large_png_files > 0:
            uniq_fileroots = numpy.sort(numpy.unique(
                numpy.concatenate([all_thumb_png_files_froots,
                                   all_large_png_files_froots])))
        elif n_thumb_png_files > 0:
            uniq_fileroots = numpy.sort(numpy.unique(
                all_thumb_png_files_froots))
        elif n_large_png_files > 0:
            uniq_fileroots = numpy.sort(numpy.unique(
                all_large_png_files_froots))
        else:
            raise IOError("No thumb-size or large-size files found."
                          "  Looking in " + orig_dir)

        # How many thumbnails per row?
        n_thumbs_per_row = 8

        # Make the HTML file for the thumbnail-sized preview plots.
        with open(odir+ofile+"_thumbs.html", 'w') as ofi:
            ofi.write('<html><head></head><body>\n')
            ofi.write('<table style="border:1px solid black;border-collapse:'
                      'collapse;width:100%">\n')

            for i, ufr in enumerate(uniq_fileroots):

                # Check if the thumb version of the preview exists.
                if n_thumb_png_files > 0 and ufr in all_thumb_png_files_froots:
                    has_thumb = True
                    where_this_thumb = numpy.where(
                        all_thumb_png_files_froots == ufr)[0]
                    if len(where_this_thumb) > 1:
                        print("Warning: This unique file root appeared more"
                              "than once in the list of all thumbnails: " +
                              ufr)
                    where_this_thumb = where_this_thumb[0]
                else:
                    has_thumb = False

                if has_thumb:
                    if i % n_thumbs_per_row == 0:
                        ofi.write('  <tr>\n')

                    # Write the cell containing the thumbnail preview, (or
                    # just fill it grey if missing).
                    ofi.write('    <td style="border:1px solid black;width:'
                              '135px;vertical-align:top"><div style="width:'
                              '128px;text-align:center"><span style='
                              '"font-weight:bold">'+'_<br>'.join(ufr.split('_'))
                              +'</span></div>')
                    if has_thumb:
                        ofi.write('    <img src="'+os.path.relpath(
                            all_thumb_png_files[where_this_thumb], odir)+
                                  '" width="128px">')
                    else:
                        ofi.write('    <div style="background-color:#86867D;'
                                  'width:128px;height:128px"></div>')
                    ofi.write('    </td>\n')

                    if i % n_thumbs_per_row == n_thumbs_per_row-1:
                        ofi.write('  </tr>\n')

                else:
                    print("Warning: Could not find full-size PNG for"
                          " IPPPSSOOT_filetype = " + ufr)
            ofi.write('</table>/n')
            ofi.write('</body></html>/n')

        # Make the HTML file for the large-sized preview plots.
        with open(odir+ofile+"_large.html", 'w') as ofi:
            ofi.write('<html><head></head><body>\n')
            ofi.write('<table style="border:1px solid black;border-collapse:'
                      'collapse;width:'+str(int(round(2.*plot_display_width)))+
                      'px">\n')
            ofi.write('<tr><th>New Version</th><th>Old Version</th></tr>\n')


            for i, ufr in enumerate(uniq_fileroots):

                # Check if the large version of the preview exists.
                if n_large_png_files > 0 and ufr in all_large_png_files_froots:
                    has_large = True
                    where_this_large = numpy.where(
                        all_large_png_files_froots == ufr)[0]
                    if len(where_this_large) > 1:
                        print("Warning: This unique file root appeared more"
                              " than once in the list of all large previews: "+
                              ufr)
                    where_this_large = where_this_large[0]
                else:
                    has_large = False

                # If it has at least one preview image, then print this table
                # row.
                if has_large:
                    ofi.write('  <tr>\n')

                    # Write the cell containing the large preview, (or just
                    # fill it grey if missing).
                    ofi.write('    <td style="border:1px solid black;width:'
                              '100%">')
                    if has_large:
                        ofi.write('    <img src="'+os.path.relpath(
                            all_large_png_files[where_this_large], odir)+
                                  '" width="'+str(plot_display_width)+'px">')
                    else:
                        ofi.write('    <div style="background-color:#86867D;'
                                  'width:'+str(plot_display_width)+'px;height:'+
                                  str(plot_display_width)+'px"></div>')
                    ofi.write('    </td>\n')

                    # If requested, write the cell containing the original
                    # preview plot.  If requested but missing, just fill it with
                    # grey.
                    if orig_dir is not None:
                        # Check if this file root has a match in the original
                        # directory.
                        orig_preview_index = find_orig_preview(ufr,
                                                               all_orig_files)
                        ofi.write('    <td style="border:1px solid black;width:'
                                  '100%">')
                        if orig_preview_index is not None:
                            ofi.write('    <img src="'+os.path.relpath(
                                all_orig_files_withpath[orig_preview_index],
                                odir)+'" width="'+str(plot_display_width)+
                                      'px">')
                        else:
                            ofi.write('    <div style="background-color:'
                                      '#86867D;width:'+str(plot_display_width)+
                                      'px;height:'+str(plot_display_width)+
                                      'px"></div>')
                        ofi.write('    </td>\n')

                    ofi.write('  </tr>\n')

                else:
                    print("Warning: Could not find full-size PNG for "
                          "IPPPSSOOT_filetype = " + ufr)

            ofi.write('</table>\n')
            ofi.write('</body></html>\n')

    else:
        raise ValueError("No preview plot directory specified.")

#--------------------

if __name__ == "__main__":

    # Create argument parser.
    PARSER = argparse.ArgumentParser(description="Create HTML page of preview"
                                     " plots, given a directory containing the"
                                     " plots.")

    PARSER.add_argument("input_dir", action="store", type=str, help="[Required]"
                        " Full path to directory containing preview plots.")

    PARSER.add_argument("-f", action="store", type=str, dest="output_file",
                        default="plot_previews", help="[Optional] Base file"
                        " name of the output HTML files.  If the files already"
                        " exist, they will be overwritten.  Defualt ="
                        " '%(default)s'.")

    PARSER.add_argument("-o", action="store", type=str, dest="output_dir",
                        default="html/plot_previews/", help="[Optional] Specify"
                        " the full path to the output HTML files.  Defualt ="
                        " '%(default)s'.")

    PARSER.add_argument("-p", action="store", type=str, dest="orig_dir",
                        help="[Optional] Full path to a directory containing"
                        " the original versions of the plot previews from CADC."
                        "  If provided, these will be included in the HTML"
                        " table for comparision purposes.")

    PARSER.add_argument("-w", action="store", type=float, dest=
                        "plot_display_width", default=512., help="[Optional]"
                        " Specify the display width of the large-sized preview"
                        " plots, in pixels, for the HTML table.  Default = "
                        "'%(default)s.'")

    # Parse arguments.
    ARGS = PARSER.parse_args()

    # Make sure the requested display width for the large-sized plots is at
    # least greater than a minimum value.
    MIN_DISPLAY_WIDTH = 128.
    if ARGS.plot_display_width < MIN_DISPLAY_WIDTH:
        print("Warning: Display width for full-size preview plots is very,"
              " very small.  Will use a display width of "+
              str(MIN_DISPLAY_WIDTH)+" px instead of "+
              str(ARGS.plot_display_width)+" px.")
        ARGS.plot_display_width = MIN_DISPLAY_WIDTH

    # Call main method.
    make_html(ARGS.input_dir, ARGS.output_dir, ARGS.output_file, ARGS.orig_dir,
              ARGS.plot_display_width)
#--------------------



#!/usr/bin/env python

"""
.. module:: make_jwst_spec_previews
   :synopsis: Given a JWST spectrum file name, this script will read in the
       data and generate preview plots of the spectra.  The plots are generated
       in different dimensions (large, thumbnail) and, depending on the
       instrument / configuration, are plotted in different formats to maximize
       readability and usability.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
from __future__ import print_function
import argparse
from os import path
from builtins import str
#--------------------
# External Imports
#--------------------
from astropy.io import fits
#--------------------
# Package Imports
#--------------------
from spec_plots.utils.specutils_jwst import readspec, plotspec
from spec_plots.utils.specutils.calc_plot_metrics import calc_plot_metrics
from spec_plots import __version__

FLUX_SCALE_FACTOR_DEFAULT = 10.
FLUXERR_SCALE_FACTOR_DEFAULT = 5.
N_CONSECUTIVE_DEFAULT = 20
OUTPUT_PATH_DEFAULT = ""
OUTPUT_TYPE_DEFAULT = str("png")
DPI_VAL_DEFAULT = 96.
DEBUG_DEFAULT = False
FULL_YLABELS_DEFAULT = False
NOOPTIMIZE_DEFAULT = False
VERBOSE_DEFAULT = False

#--------------------

class JWSTSpecPrevError(Exception, object):
    """
    This class defines a generic Exception to use for errors raised in
    MAKE_JWST_SPEC_PREVIEWS.  It simply prints the given string when raising the
    exception. e.g.,

    .. code-block:: python

         raise JWSTSpecPrevError("Print a string")
         JWSTSpecPrevError: *** MAKE_JWST_SPEC_PREVIEWS ERROR: 'Print a string'
    """

    def __init__(self, value):
        """
        Initiate the Exception.

        :param value: The string to print on error.

        :type value: str
        """
        super(JWSTSpecPrevError, self).__init__(value)
        self.value = value

    def __str__(self):
        """
        Overrides the str function for this class.
        """
        return "*** MAKE_JWST_SPEC_PREVIEWS ERROR: "+repr(self.value)

#--------------------

def check_input_options(args):
    """
    Check that input arguments satisfy some minimum requirements.

    :param args: Stores arguments and options.

    :type args: argparse.Namespace object.

    :raises: JWSTSpecPrevError, ValueError
    """

    # Make sure the input file is specified on input (non-empty string), and
    # remove any leading/trailing whitespace.  Catch as an ArgumentsParser error
    # in this case.  If it does exist, check that the input file exists at the
    # time of the command-line parameter checking.
    if not args.input_file:
        raise JWSTSpecPrevError("File name must be specified.")
    else:
        args.input_file = args.input_file.strip()

    # Make sure the output_type string is trimmed and lowercase.
    args.output_type = args.output_type.strip().lower()

    # The DPI value must be greater than zero...
    if args.dpi_val <= 0.:
        raise ValueError("DPI value must be > 0.")

#--------------------

def get_instrument_name(input_file):
    """
    Retrieves the instrument name from a FITS file primary header.

    :param input_file: Name of input FITS file.

    :type input_file: str

    :returns: str -- The name of the instrument based on the FITS header
        keyword, in uppercase with leading/trailing whitespace removed.

    :raises: KeyError, JWSTSpecPrevError
    """

    # Make sure file exists.
    if not path.isfile(input_file):
        raise JWSTSpecPrevError("Input file not found, looking for file: " +
                                input_file)

    with fits.open(input_file) as hdulist:
        # Make sure the INSTRUME keyword exists in the primary header,
        # otherwise, catch as a KeyError in this case.
        try:
            this_instrument = hdulist[0].header["INSTRUME"]
        except KeyError:
            print("*** MAKE_JWST_SPEC_PREVIEWS ERROR: INSTRUME keyword not"
                  " found in this file's primary header: " + input_file)
            exit(1)
    return this_instrument.strip().upper()

#--------------------

def make_jwst_spec_previews(input_file, flux_scale_factor=
                            FLUX_SCALE_FACTOR_DEFAULT, fluxerr_scale_factor=
                            FLUXERR_SCALE_FACTOR_DEFAULT, n_consecutive=
                            N_CONSECUTIVE_DEFAULT,
                            output_path=OUTPUT_PATH_DEFAULT,
                            output_type=OUTPUT_TYPE_DEFAULT,
                            dpi_val=DPI_VAL_DEFAULT, debug=DEBUG_DEFAULT,
                            full_ylabels=FULL_YLABELS_DEFAULT, optimize=
                            not NOOPTIMIZE_DEFAULT, verbose=VERBOSE_DEFAULT):
    """
    Main function in the module.

    :param input_file: The full path and name of the FITS file to create a
        preview for.

    :type input_file: str

    :param flux_scale_factor: The ratio between the flux and the median flux
        that defines the pass/fail criterion within the edge trim test.

    :type flux_scale_factor: float

    :param fluxerr_scale_factor: The ratio between the flux uncertainty and the
        median flux uncertainty that defines the pass/fail criterion within the
        edge trim test.

    :type fluxerr_scale_factor: float

    :param n_consecutive: The number of consecutive data points that must pass
        the edge trim test to define the start and end of the spectrum for
        plotting purposes.

    :type n_consecutive: int

    :param output_path: Full path to output plot files.  Do not include the
        output file name in this path.  If not supplied, plots will be created
        in the same directory as the input file.

    :type output_path: str

    :param output_type: The file type of the plots to make.

    :type output_type: str

    :param dpi_val: The DPI value of your device's monitor, which will affect
        the size of the output plots.

    :type dpi_val: float

    :param debug: If True, turns on debug mode, which will plot to the screen
        and color-code fluxes based on different rejection criteria.

    :type debug: bool

    :param full_ylabels: If True, label the y-axis with full values, including
        powers of ten in scientific notation.

    :type full_ylabels: bool

    :param optimize: If set to True, will use a slightly optimized version of
        determining the plot covering fraction.

    :type optimize: bool

    :param verbose: Turn on verbose messages/logging?

    :type verbose: bool

    :raises: JWSTSpecPrevError
    """

    # Print file name, if verbose is turned on.
    if verbose:
        print("Input file: " + input_file)

    # Derive output file name from input file name.
    if output_type != "screen":
        output_file = (path.join(output_path, "") +
                       path.basename(input_file).split(".fits")[0] + "." +
                       output_type)
    else:
        output_file = None

    # Print name of output file, if verbose is turned on and not plotting to
    # screen.
    if verbose and output_type != "screen":
        print("Output file: " + output_file)
    elif verbose:
        print("Output file: Plotting to screen.")

    # Read in the FITS file to determine which instrument it comes from.
    # Print the name of the instrument found in the header if verbose is turned
    # on.
    this_instrument = get_instrument_name(input_file)
    if verbose:
        print("Instrument: " + this_instrument)

    # Read in the FITS files and create plots using the local package
    # appropriate for the instrument used in the input file.
    if this_instrument in ["MIRI", "NIRSPEC", "NIRISS"]:
        # Get wavelengths, fluxes, flux uncertainties.
        jwst_spectrum = readspec(input_file)

        # Calculate plot metrics.
        spec_plot_metrics = calc_plot_metrics(this_instrument.lower(),
                                              jwst_spectrum.wavelengths,
                                              jwst_spectrum.fluxes,
                                              jwst_spectrum.fluxerrs,
                                              jwst_spectrum.dqs,
                                              n_consecutive, flux_scale_factor,
                                              fluxerr_scale_factor)

        # Make "large-size" plot.
        plotspec(jwst_spectrum, output_type, output_file,
                 flux_scale_factor,
                 fluxerr_scale_factor, spec_plot_metrics,
                 dpi_val=dpi_val, output_size=1024, debug=debug,
                 full_ylabels=full_ylabels,
                 optimize=optimize)

        if not debug:
            # Make "thumbnail-size" plot, if requested.
            plotspec(jwst_spectrum, output_type, output_file,
                     flux_scale_factor,
                     fluxerr_scale_factor, spec_plot_metrics,
                     dpi_val=dpi_val, output_size=128,
                     optimize=optimize)
    else:
        raise JWSTSpecPrevError("'INSTRUME' keyword not understood: " +
                                this_instrument)

#--------------------

def setup_args():
    """
    Set up command-line arguments and options.

    :returns: ArgumentParser -- Stores arguments and options.
    """
    parser = argparse.ArgumentParser(
        description="Create spectroscopic preview plots given an JWST spectrum"
        " FITS file.")

    parser.add_argument("input_file", action="store", type=str, help=
                        "[Required] Full path to input file (JWST spectrum) for"
                        " which to generate preview plots.  Include the file"
                        " name in the path.")

    parser.add_argument("-d", action="store_true", dest="debug", default=
                        DEBUG_DEFAULT, help="[Optional] Turn on debug mode,"
                        " which will plot to the screen and color-code fluxes"
                        " based on different rejection criteria.  Default ="
                        " %(default)s.")

    parser.add_argument("--dpival", action="store", type=float, dest="dpi_val",
                        default=DPI_VAL_DEFAULT, help="[Optional] Specify the"
                        " DPI value of your device's monitor, which will affect"
                        " the size of the output plots.  Default ="
                        " %(default)s.")

    parser.add_argument("-e", action="store", type=float, dest=
                        "fluxerr_scale_factor", default=
                        FLUXERR_SCALE_FACTOR_DEFAULT, help="[Optional] Specify"
                        " the ratio between the flux uncertainty and the median"
                        " flux uncertainty that defines the pass/fail criterion"
                        " within the edge trim test.  Default = %(default)s.")

    parser.add_argument("-n", action="store", type=int, dest="n_consecutive",
                        default=N_CONSECUTIVE_DEFAULT, help="[Optional] Specify"
                        " the number of consecutive data points that must pass"
                        " the edge trim test to define the start and end of the"
                        " spectrum for plotting purposes.  Default ="
                        " %(default)i.")

    parser.add_argument("--nooptimize", action="store_true", dest="nooptimize",
                        default=NOOPTIMIZE_DEFAULT, help="[Optional] If True,"
                        " do not use an optimized method of calculating plot"
                        " covering fraction to determine line transparency."
                        "  Default = %(default)s.")

    parser.add_argument("-o", action="store", type=str, dest="output_path",
                        default=OUTPUT_PATH_DEFAULT, help="[Optional] Full path"
                        " to output plot files.  Do not include file name in"
                        " path.    Default = %(default)s, the same directory as"
                        " the input file.")

    parser.add_argument("-s", action="store", type=float, dest=
                        "flux_scale_factor", default=FLUX_SCALE_FACTOR_DEFAULT,
                        help="[Optional] Specify the ratio between the flux and"
                        " the median flux that defines the pass/fail criterion"
                        " within the edge trim test.  Default = %(default)s.")

    parser.add_argument("-t", action="store", type=str.lower, dest=
                        "output_type", default=OUTPUT_TYPE_DEFAULT, help=
                        "[Optional] Specify the file type of the plots to make."
                        "  Default = %(default)s.", choices=['png', 'eps',
                                                             'screen'],
                        metavar="{png,ps,screen}")

    parser.add_argument("-v", action="store_true", dest="verbose",
                        default=VERBOSE_DEFAULT, help="[Optional] Turn on"
                        " verbose messages/logging?  Default = %(default)s.")

    parser.add_argument("-y", action="store_true", dest="full_ylabels",
                        default=FULL_YLABELS_DEFAULT, help="[Optional] Label"
                        " the y-axis with full values, including powers of ten"
                        " in scientific notation?  Default = %(default)s.")

    return parser

#--------------------

if __name__ == "__main__":

    # Create ArgumentParser object that holds arguments and options.
    INPUT_ARGS = setup_args().parse_args()

    # Check arguments and options.
    check_input_options(INPUT_ARGS)

    # Call main function.
    make_jwst_spec_previews(INPUT_ARGS.input_file,
                            flux_scale_factor=INPUT_ARGS.flux_scale_factor,
                            fluxerr_scale_factor=(
                                INPUT_ARGS.fluxerr_scale_factor),
                            n_consecutive=INPUT_ARGS.n_consecutive,
                            output_path=INPUT_ARGS.output_path,
                            output_type=INPUT_ARGS.output_type,
                            dpi_val=INPUT_ARGS.dpi_val,
                            debug=INPUT_ARGS.debug,
                            full_ylabels=INPUT_ARGS.full_ylabels,
                            optimize=not INPUT_ARGS.nooptimize,
                            verbose=INPUT_ARGS.verbose)
#--------------------



"""
.. module:: setup

   :synopsis: This script is used to setup the pip packages.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

from __future__ import absolute_import
from setuptools import setup
from spec_plots import __version__

setup(name="spec_plots",
      version=__version__,
      description="Create preview plots of HST or JWST spectra.",
      classifiers=["Programming Language :: Python :: 2.7",
                   "Programming Language :: Python :: 3.5"],
      url="https://github.com/openSAIL/spec_plots",
      author="Scott W. Fleming",
      author_email="fleming@stsci.edu",
      license="MIT",
      packages=["spec_plots", "spec_plots.utils", "spec_plots.utils.specutils",
                "spec_plots.utils.specutils_cos",
                "spec_plots.utils.specutils_jwst",
                "spec_plots.utils.specutils_stis"],
      install_requires=["astropy>=0.4.1", "matplotlib>=1.4.1", "numpy>=1.9.1",
                        "future>=0.15"],
      entry_points={"console_scripts" :
                    ["make_hst_spec_previews = spec_plots.__main__:main",
                     "make_jwst_spec_previews = spec_plots.__main_jwst__:main"
                    ]},
      zip_safe=False)
